"""
QLauncher
================

Program made for easier usage of Quantum computer for testing purposes.
Provided with pre-made problems and algorithms.
"""
from qlauncher.base import Result
from qlauncher.launcher import QLauncher
